/*
 * Copyright 2014 Cloud Linux Zug GmbH
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 *
 * The part the support library for lsapi & proxy_lsapi Apache modules
 * author Alexander Demeshko <ademeshko@cloudlinux.com>
 *
 */
 
#include <lscapi_util.h>

#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#include <mod_core.h>

//#pragma GCC visibility push(default)

const char* lscapi_php_value_handler(cmd_parms* cmd, void *cfg,
                                    const char *name, const char *value,
                                    int isAdmin) {
    php_param_t *eptr = apr_palloc(cmd->pool, sizeof(php_param_t));

    if(!strncasecmp(value, "none", 4)) {
        value = "";
    }

    eptr->valueLen = strlen(value);
    eptr->value = apr_pmemdup(cmd->pool, value, eptr->valueLen+1);
    eptr->isAdmin = isAdmin;
	//eptr->htaccess = ((cmd->override & (RSRC_CONF|ACCESS_CONF)) == 0);

	//ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, cmd->server, LSCAPI_APLOGNO
    //             "LSCAPI_PhpValueHandler: (%s)->(%s) isAdmin:%d", name, eptr->value, eptr->isAdmin);
    apr_hash_t* phpParams = ((lsapi_dir_conf_t*)cfg)->phpParams;

    apr_hash_set(phpParams, name, APR_HASH_KEY_STRING, eptr);

    return NULL;
}

const char *lscapi_php_flag_handler(cmd_parms *cmd, void *cfg,
                                    const char *name, const char *value, int isAdmin) {
    char boolVal[2];

    boolVal[0] = (!strcasecmp(value, "On") || (value[0] == '1' && value[1] == '\0')) ? '1' : '0';
	boolVal[1] = '\0';
    return lscapi_php_value_handler(cmd, cfg, name, boolVal, isAdmin);
}

void *lscapi_create_dir_config(apr_pool_t *pool, char *x) {
    lsapi_dir_conf_t *cfg = apr_palloc(pool, sizeof(lsapi_dir_conf_t));
    cfg->phpParams = apr_hash_make(pool);
    return cfg;
}

void *lscapi_merge_dir_config(apr_pool_t *pool, void *BASE, void *CUR) {
    lsapi_dir_conf_t *base = BASE;
    lsapi_dir_conf_t *cur = CUR;
    apr_hash_t *curH = cur->phpParams;
    apr_hash_t *baseH = base->phpParams;

    // initially lets create new config as a copy of the base one
    apr_hash_t *newH = apr_hash_copy(pool, baseH);

    // iterate over current config to check its values against values of the base config
    for(apr_hash_index_t *hi = apr_hash_first(pool, curH); hi; hi = apr_hash_next(hi)) {
        const char *key;
        apr_ssize_t keyLen;
        php_param_t *curVal;
        apr_hash_this(hi, (const void**)(&key), &keyLen, (void**)(&curVal));

        if(curVal->isAdmin) { // new admin value will beat even old admin one
            apr_hash_set(newH, key, keyLen, curVal);
            continue;
        }

        php_param_t *baseVal = apr_hash_get(baseH, key, keyLen);
        if(!baseVal                  // there is no such key in the base config
           || !baseVal->isAdmin)  {  // old value is not admin so new one will beat it

            apr_hash_set(newH, key, keyLen, curVal);
        }

    } // for


    lsapi_dir_conf_t *cfg = apr_palloc(pool, sizeof(lsapi_dir_conf_t));
    cfg->phpParams = newH;

    return cfg;
}

void *lscapi_create_svr_config(apr_pool_t *pool, server_rec *s) {
    //ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, LSCAPI_APLOGNO
    //                  "entering lsapi_create_svr_config" );
    lsapi_svr_conf_t *cfg = apr_pcalloc(pool, sizeof(lsapi_svr_conf_t));
#if MODULE_MAGIC_NUMBER_MAJOR > 20081212
    cfg->default_uid = ap_unixd_config.user_id;
    cfg->default_gid = ap_unixd_config.group_id;
#else
    cfg->default_uid = unixd_config.user_id;
    cfg->default_gid = unixd_config.group_id;
#endif
    cfg->auth_by_uidgid = apr_hash_make(pool);
    cfg->terminate_backends_on_exit = 1;
    
    lscapi_init_backend_info(&cfg->backend_info);
    return cfg;
}

void *lscapi_merge_svr_config(apr_pool_t *pool, void *BASE, void *CUR) {
    lsapi_svr_conf_t *base = BASE;
    lsapi_svr_conf_t *cur = CUR;

    lsapi_svr_conf_t *cfg = apr_palloc(pool, sizeof(lsapi_svr_conf_t));
    cfg->default_uid = cur->default_uid;
    cfg->default_gid = cur->default_gid;
    cfg->auth_by_uidgid = cur->auth_by_uidgid;
    
    if(cur->secret_found) {
        cfg->secret_found = 1;
        memcpy(cfg->secret, cur->secret, sizeof cfg->secret);
    } else if(base->secret_found) {
        cfg->secret_found = 1;
        memcpy(cfg->secret, base->secret, sizeof cfg->secret);
    }
    
    if(cur->terminate_backends_was_set) {
        cfg->terminate_backends_on_exit = cur->terminate_backends_on_exit;
    } else if(base->terminate_backends_was_set) {
        cfg->terminate_backends_on_exit = base->terminate_backends_on_exit;
    }
    
    if(cur->debug_enabled_was_set) {
        cfg->debug_enabled = cur->debug_enabled;
    } else if(base->debug_enabled_was_set) {
        cfg->debug_enabled = base->debug_enabled;
    }

    lscapi_merge_backend_info(&base->backend_info, &cur->backend_info, &cfg->backend_info);

    return cfg;
}


const char *lscapi_secret_file_handler(cmd_parms *cmd, lsapi_svr_conf_t *cfg, const char *value) 
{

    //ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, cmd->server, LSCAPI_APLOGNO
    //                  "entering lsapi_secret_file_handler(%s)", value );
    char errbuf[256];
    if(lscapi_read_secret(value, cfg->secret, errbuf, sizeof errbuf) != 0) {
        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, cmd->server, LSCAPI_APLOGNO
                      "ReadSecret(%s) failed: %s", value, errbuf );
        return NULL;
    }

    cfg->secret_found = 1;
/*
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, cmd->server, LSCAPI_APLOGNO
                      "lsapi_secret_file_handler: secret is %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x ",
                      cfg->secret[0], cfg->secret[1], cfg->secret[2], cfg->secret[3],
                      cfg->secret[4], cfg->secret[5], cfg->secret[6], cfg->secret[7],
                      cfg->secret[8], cfg->secret[9], cfg->secret[10], cfg->secret[11],
                      cfg->secret[12], cfg->secret[13], cfg->secret[14], cfg->secret[15]);
*/
    return NULL;
}


const char *lscapi_suphp_usergroup_handler(cmd_parms *cmd, lsapi_svr_conf_t *cfg,
                                           const char *arg1, const char *arg2)
{
    struct passwd *pw = getpwnam(arg1);
    if(!pw) {
        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, cmd->server, "suPHP_UserGroup: Wrong user(%s) - ignore directive", arg1);
        return NULL;
    }
    
    struct group *grp = getgrnam(arg2);
    if(!grp) {
        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, cmd->server, "suPHP_UserGroup: Wrong group(%s) - ignore directive", arg2);
        return NULL;
    }

    cfg->suphp_uid = pw->pw_uid;
    cfg->suphp_gid = grp->gr_gid;

    return NULL;
}


auth_by_uidgid_t* lscapi_get_signed_suexec(request_rec *r, const module *m) 
{
    ap_unix_identity_t *ugid = ap_run_get_suexec_identity(r);
    if(!ugid) {
        //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, conn->r, LSCAPI_APLOGNO
        //            "suEXEC info not found");
        return NULL;
    }

    //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, conn->r, LSCAPI_APLOGNO
    //                "suEXEC info found: uid-%d; gid-%d", ugid->uid, ugid->gid);


    lsapi_svr_conf_t *cfg = ap_get_module_config(r->server->module_config, m);
    if(!cfg || !cfg->secret_found) {
        //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, conn->r, LSCAPI_APLOGNO
        //            "secret file not found or not configured");
        return NULL;
    }

    auth_by_uidgid_t *auth = apr_hash_get(cfg->auth_by_uidgid, ugid, sizeof(ap_unix_identity_t) );
    if(!auth) {
        //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, conn->r, LSCAPI_APLOGNO
        //            "The hash does not contain our entry - trying to create and insert");
        auth = apr_palloc(r->server->process->pool, sizeof(auth_by_uidgid_t));
        auth->uidgid[0] = ugid->uid;
        auth->uidgid[1] = ugid->gid;
        lscapi_sign_uidgid(cfg->secret, auth->uidgid, auth->sign);

        apr_hash_set(cfg->auth_by_uidgid, ugid, sizeof(ap_unix_identity_t), auth );
    //} else {
    //    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, conn->r, LSCAPI_APLOGNO
    //                "The hash contains our entry");
    }
    return auth;
}

//#pragma GCC visibility pop
